-- by mor2000
-- v2.0
-- PlaceableHeaps

source(g_currentModDirectory.."HeapShovelFillTrigger.lua");

PlaceableHeap = {};
InitObjectClass(PlaceableHeap, "PlaceableHeap");

function PlaceableHeap:new(isServer, isClient, mt)
	--print("new Heap");
	if PlaceableHeap_mt == nil then
		PlaceableHeap_mt = Class(PlaceableHeap, Placeable);
	end;
	
	if mt == nil then
		mt = PlaceableHeap_mt;
	end;

    self = Placeable:new(isServer, isClient, mt);
    registerObjectClassName(self, "PlaceableHeap");
	
	self.isServer=isServer;
	self.isClient=isClient;
	self.mt=mt;

	self.nodeId=0;

	self.update=PlaceableHeap.update;

	self.setFillLevel=PlaceableHeap.setFillLevel;
	self.getFillLevel=PlaceableHeap.getFillLevel;

	self.getAcceptedFillTypes=PlaceableHeap.getAcceptedFillTypes;
	self.setFillType=PlaceableHeap.setFillType;
	self.getFillType=PlaceableHeap.getFillType;
	
	self.getTipDistanceFromTrailer=PlaceableHeap.getTipDistanceFromTrailer;
	self.getTipDistance=PlaceableHeap.getTipDistance;
	self.getNoAllowedText=TipTrigger.getNoAllowedText;
	self.tipTriggerCallback=PlaceableHeap.tipTriggerCallback;
	self.updateTrailerTipping=PlaceableHeap.updateTrailerTipping;
	self.updateMoving=function() end;
	
	self.getAllowShovelFillType = PlaceableHeap.getAllowShovelFillType;
	self.addShovelFillLevel = PlaceableHeap.addShovelFillLevel;

	self.displayTriggerCallback = GreenhousePlaceable.onWaterTankTrigger;
	self.getShowInfo = GreenhousePlaceable.getShowInfo;
	self.addWaterTrailer=function(trailer) end;
	self.removeWaterTrailer=function(trailer) end;

	self.findTrailerRaycastCallback=Vehicle.findTrailerRaycastCallback;
	
	self.allowFillType=PlaceableHeap.allowFillType;
	self.getAllowFillFromAir=PlaceableHeap.getAllowFillFromAir;
	self.resetFillLevelIfNeeded=PlaceableHeap.resetFillLevelIfNeeded;
	self.getIsAttachedTo=PlaceableHeap.getIsAttachedTo;
	
	self.tipTrigger=nil;
	self.shovelTrigger=nil;
	
	self.triggerId=nil;
	self.fillRootNode=nil;
	self.rootNode = nil;
	self.exactFillRootNode=nil;
	self.allowFillFromAir=true;
	
	self.fillLevel=0;
	self.sentFillLevel=self.fillLevel;
	self.fillType=0;
	self.sentFillType=self.fillType;
	self.capacity=100000;
	
	self.acceptedFillTypes={};
	self.isDeleted=false;
	self.fillAutoAimTargetNode=nil;
	
	self.heapPlanes={};
	self.activePlaneId=0;
	self.activePlaneShovelTriggerId=0;
	self.activePlaneShovelTriggerOn=false;
	self.activePlaneX=0;
	self.activePlaneY=nil;
	self.activePlaneZ=0;
	self.moveMaxAmount=700000;
	self.moveMaxY=0;
	self.moveMinY=-3.75;
	
	self.getHemisphereRadiusFromVolume=PlaceableHeap.getHemisphereRadiusFromVolume;
	self.hemisphereRadiusMaxAmount=0;

	self.forceMovePlane = PlaceableHeap.forceMovePlane;
	self.activePlaneRepostioned=false;
	
	self.isAreaTrigger = false;
	
	self.tipTriggerId=nil;
	self.triggerStartId = nil;
	self.triggerTipWidth=nil;
	self.isEnabled = true;
	
	self.displayTriggerId=nil;
	self.playerInRange=false;
	self.vehiclesInRange = {};
	
	self.showDumpIcon=true;
	self.dumpIconId=0;
	self.fruitTypePlacement=64;
	
	self.placeableHeapDirtyFlag = self:getNextDirtyFlag();
	
	self.timediff=0;
	--print("Heap created");
	return self;
end;

function PlaceableHeap:delete()
	--BaseMission:removeVehicle(self, false);

	if self.tipTriggerId~=nil and self.tipTriggerId~=0 then
		removeTrigger(self.tipTriggerId);
	end;

	if(self.shovelTrigger~=nil and self.shovelTrigger~=0)then
		self.shovelTrigger:delete();
	end;

	if(self.displayTriggerId~=0)then
		removeTrigger(self.displayTriggerId);
	end;
	
	unregisterObjectClassName(self);
    PlaceableHeap:superClass().delete(self);
end;

function PlaceableHeap:deleteFinal()
    PlaceableHeap:superClass().deleteFinal(self);
end;

-- Verbindung initialisieren (?)

function PlaceableHeap:readStream(streamId, connection)
PlaceableHeap:superClass().readStream(self, streamId, connection);
  if connection:getIsServer() then
	self:setFillLevel(streamReadFloat32(streamId));
	self:setFillType(streamReadInt8(streamId));
  end;
end;

function PlaceableHeap:writeStream(streamId, connection)
  PlaceableHeap:superClass().writeStream(self, streamId, connection);
  if not connection:getIsServer() then
	streamWriteFloat32(streamId, self:getFillLevel());
    streamWriteInt8(streamId, self:getFillType());
  end;
end;

-- Verbindung akutalisieren (?)

function PlaceableHeap:readUpdateStream(streamId, timestamp, connection)
  PlaceableHeap:superClass().readUpdateStream(self, streamId, timestamp, connection);
  if connection:getIsServer() then
	self:setFillLevel(streamReadFloat32(streamId));
	self:setFillType(streamReadInt8(streamId));
  end;
end;

function PlaceableHeap:writeUpdateStream(streamId, connection, dirtyMask)
  PlaceableHeap:superClass().writeUpdateStream(self, streamId, connection, dirtyMask);
  if not connection:getIsServer() then
	streamWriteFloat32(streamId, self:getFillLevel());
    streamWriteInt8(streamId, self:getFillType());
  end;
end;

function PlaceableHeap:updateTick(dt)
	if self.isServer then
		if self.fillLevel ~= self.sentFillLevel or self.fillType ~= self.sentFillType then
			self:raiseDirtyFlags(self.placeableHeapDirtyFlag);
			self.sentFillLevel = self.fillLevel;
			self.sentFillType = self.fillType;
		end;
	end;
end;

function PlaceableHeap:load(xmlFilename, x,y,z, rx,ry,rz, moveMode, initRandom)
	--print("loading Heap");
	if not PlaceableHeap:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, moveMode, initRandom) then
        return false;
    end;
	--print("loading Heap successful");
	
	--print("Filename: "..xmlFilename..", moveMode: "..moveMode);

	self.xml = xmlFilename;
	self.pos = {x=x,y=y,z=z};
	self.rot = {rx,ry,rz};
	
	if not moveMode then
		
		local xmlFile = loadXMLFile("TempXML", xmlFilename);
		
		-- set accepted FillTypes
		local fillTypesString=Utils.getNoNil(getXMLString(xmlFile, "placeable.FillTypes#names"),"");
		for k,v in pairs(Utils.splitString(" ",fillTypesString)) do
			local desc2 = Fillable.fillTypeNameToDesc[v];
			if type(desc2)=="table" then
				--print("index for "..v..": "..desc2.index);
				self.acceptedFillTypes[desc2.index] = true;
			else
				--print("index for "..v.." not found");
			end;
		end;

		--print("attempt to add Heap Trigger"); -- 1 tipTrigger expected
		--getChildren(self.nodeId); -- see structure of Ids
		local TriggerGroupId=Utils.getNoNil(getChild(self.nodeId,"Trigger"),0);
		if TriggerGroupId~=0 then
			local tipTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"tipTrigger"),0);
			if tipTriggerId ~= nil then
				addTrigger(tipTriggerId, "tipTriggerCallback", self);
				self.tipTriggerId=tipTriggerId;
			end;
			
			local shovelTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"shovelTrigger"),0);
			if shovelTriggerId ~= 0 then
				self.shovelTriggerId=shovelTriggerId;
			end;
			
			local dumpTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"dumpTrigger"),0);
			if dumpTriggerId ~= 0 then
				--print("dumpTrigger found with Id "..dumpTriggerId);
				self.fillRootNode=dumpTriggerId;
				self.rootNode = dumpTriggerId;
				self.exactFillRootNode=dumpTriggerId;
				self.fillAutoAimTargetNode=dumpTriggerId;
				g_currentMission.nodeToVehicle[dumpTriggerId] = self;
			end;

			local displayTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"displayTrigger"),0);
			if displayTriggerId ~= 0 then
				self.displayTriggerId=displayTriggerId;
				addTrigger(self.displayTriggerId, "displayTriggerCallback", self);
			end;
		end;

		--print("attempt to add Heap Planes");
		local heapPlanesId = Utils.getNoNil(getChild(self.nodeId,"HeapPlanes"),0);
		if(heapPlanesId~=0)then
			self.moveMaxAmount=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMaxAmount"), self.moveMaxAmount);
			self.moveMaxY=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMaxY"), self.moveMaxY);
			self.moveMinY=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMinY"), self.moveMinY);
			for i=1,getNumOfChildren(heapPlanesId) do
				local heapPlaneId = getChildAt(heapPlanesId, i-1);
				if(heapPlaneId~=0)then
					local fillTypeString=Utils.getNoNil(getUserAttribute(heapPlaneId, "fillType"), 0);
					local desc2=nil;
					if(fillTypeString=="placement")then
						desc2={index=self.fruitTypePlacement};
					else
						desc2 = Fillable.fillTypeNameToDesc[fillTypeString];
					end;
					if type(desc2)=="table" then
						self.heapPlanes[desc2.index]= heapPlaneId;
						local x,y,z = getTranslation(heapPlaneId)
						setTranslation(heapPlaneId,x,self.moveMinY-1,z);
						setVisibility(heapPlaneId,false);
					end;
				end;
			end;
		end;
		
		local iconId=Utils.getNoNil(getChild(self.nodeId,"dumpIcon"),0);
		if(iconId~=0)then
			self.dumpIconId=iconId;
		end;

		self.triggerStartId = self.nodeId;
		self.triggerTipWidth=Utils.getNoNil(getUserAttribute(id, "triggerTipWidth"), 12);

	end;

	self:setFillLevel(0);
	self.sentFillLevel=self.fillLevel;
	self:setFillType(0);
	self.sentFillType=self.fillType;
	
	--print("loading Heap finished");
	return true;
end;

function PlaceableHeap:update(dt)
	--print(self:getFillLevel());
	
	if(self.timediff<1000)then
		self.timediff=self.timediff+dt;
	end;
	
	if self:getShowInfo() then
    g_currentMission:addExtraPrintText(g_i18n:getText("fill_level").." "..math.floor(self:getFillLevel()));
	end
	--self.timeSinceLastUpdate=self.timeSinceLastUpdate+dt;
	if(self.fillLevel==0 and g_gui.currentGuiName=="PlacementScreen")then
		setTranslation(self.heapPlanes[self.fruitTypePlacement],0,self.moveMinY+0.5,0);
	elseif(self.fillLevel==0 and g_gui.currentGuiName=="ShopScreen")then
		setTranslation(self.heapPlanes[self.fruitTypePlacement],0,self.moveMinY-1,0);
	elseif(self.activePlaneRepostioned)then
		if(self.activePlaneId~=0)then
			setTranslation(self.activePlaneId,self.activePlaneX,self.activePlaneY,self.activePlaneZ);
		end;
		setTranslation(self.shovelTriggerId,self.activePlaneX,self.activePlaneY+0.5,self.activePlaneZ);
		self.activePlaneRepostioned=false;
	end;

	if(self.showDumpIcon and self.dumpIconId~=0)then
		rotate(self.dumpIconId, 0, 0.002 * dt, 0);
	end;

	if(self.timediff>1000 and g_currentMission:getIsServer())then
		if g_currentMission.controlPlayer and self.playerInRange and self.fillLevel==0 then
	    	g_currentMission:addHelpButtonText(g_i18n:getText("PlaceableHeapSell"), InputBinding.PLACEABLEHEAP_SELL);
			if InputBinding.isPressed(InputBinding.PLACEABLEHEAP_SELL) then
				g_client:getServerConnection():sendEvent(SellPlaceableEvent:new(self));
			end;
		end;
	end;
end;

function PlaceableHeap:getAllowShovelFillType(fillType)
	--print("Ask to fill from Shovel of Type "..tostring(fillType));
	if(self.fillType==0)then
		return self.acceptedFillTypes[fillType];
	else
		return self.fillType==fillType;
	end;
end;

function PlaceableHeap:addShovelFillLevel(shovel, fillLevelDelta, fillType)
	--print("addShovelFillLevel with "..tostring(fillLevelDelta));
	if self:getAllowShovelFillType(fillType) then
		self:setFillType(fillType);
		local oldFillLevel = self:getFillLevel();
		--print("Old FillLevel: "..tostring(oldFillLevel)..", new one: "..tostring(oldFillLevel+fillLevelDelta));
		self:setFillLevel(oldFillLevel+fillLevelDelta);
		fillLevelDelta = self:getFillLevel() - oldFillLevel;
		return fillLevelDelta;
	end;
	return 0;
end;

function PlaceableHeap:updateMoving(delta)
end;

function PlaceableHeap:getSaveAttributesAndNodes(nodeIdent)
	local attributes, nodes = PlaceableHeap:superClass().getSaveAttributesAndNodes(self, nodeIdent);
	for k,v in pairs({ fillLevel=self.fillLevel,
						fillType=self.fillType}) do
		attributes = attributes .. " "..tostring(k).."=\"" .. v .. "\"";
	end;
	return attributes, nodes
end;

function PlaceableHeap:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not PlaceableHeap:superClass().loadFromAttributesAndNodes(self, xmlFile, key, resetVehicles) then
		return false;
	end
	local fillLevel = getXMLFloat(xmlFile, key .. "#fillLevel");
	if fillLevel ~= nil then
		self:setFillLevel(fillLevel);
	end;
	local fillType = getXMLFloat(xmlFile, key .. "#fillType");
	if fillType ~= nil then
		self:setFillType(fillType);
	end;
	return true;
end;

function PlaceableHeap:getFillLevel()
	return self.fillLevel;
end;

function PlaceableHeap:setFillLevel(newFillLevel,fillType)
	local oldFillLevel = self.fillLevel;
	self.fillLevel = newFillLevel;
	
	if(fillType~=nil)then
		self:setFillType(fillType);
	end;
	
	--print("Old FillLevel: "..tostring(oldFillLevel)..", new one: "..tostring(self.fillLevel));
	
	--print("Heap FillLevel set to "..newFillLevel.." = "..tostring(self.fillLevel));

	if self.fillLevel < 0.001 then
		self.fillLevel = 0;
		self:setFillType(0);
		self.activePlaneY=self.moveMinY-1;
		self.activePlaneRepostioned=true;
	else
		if self.fillLevel ~= oldFillLevel then
			self:forceMovePlane();
		end;
	end;
	
	if(self.fillLevel<1000 and self.showDumpIcon==false)then
		setVisibility(self.dumpIconId,true);
		self.showDumpIcon=true;
	elseif(self.fillLevel>=1000 and self.showDumpIcon==true)then
		setVisibility(self.dumpIconId,false);
		self.showDumpIcon=false;
	end;
	
	self:raiseDirtyFlags(self.placeableHeapDirtyFlag);
	
	self.capacity=self.fillLevel + 100000;
end;

function PlaceableHeap:getHemisphereRadiusFromVolume(volume)
	if volume==0 then
		return 0;
	else
		return (volume/(4/6*math.pi))^(1/3);
	end;
end;

function PlaceableHeap:forceMovePlane()
	if(self.activePlaneId~=0)then
		if(self.hemisphereRadiusMaxAmount==0)then
			self.hemisphereRadiusMaxAmount=self:getHemisphereRadiusFromVolume(self.moveMaxAmount);
		end;
		local amount=math.min(self.moveMaxAmount,Utils.getNoNil(self:getFillLevel(),0));
		local ratio=self:getHemisphereRadiusFromVolume(amount)/self.hemisphereRadiusMaxAmount;
		self.activePlaneY=(self.moveMaxY-self.moveMinY)*ratio+self.moveMinY;
		self.activePlaneRepostioned=true;
	end;
end;

function PlaceableHeap:setFillType(fillType)
	if(fillType==0)then
		-- hide plane of old filltype
		if(self.shovelTrigger~=nil and self.shovelTrigger~=0)then
			self.shovelTrigger:delete();
			--print("Removed Shovel Trigger");
			self.activePlaneShovelTriggerOn=false;
		end;
		--print("Hiding Plane of old FillType");
		if self.activePlaneId~=nil and self.activePlaneId~=0 then
			setVisibility(self.activePlaneId,false);
		end;
		self.activePlaneY=self.moveMinY-1;
		self.activePlaneRepostioned=true;
	end;
	
	if(self.fillType~=fillType and fillType~=0)then
		--print("Heap has new Filltype: "..tostring(fillType));
		if self.shovelTriggerId~=nil then
			-- show plane of new fill type
			self.shovelTrigger = HeapShovelFillTrigger:onCreate(self.shovelTriggerId,self,fillType);
			--print("Added Shovel Trigger with Id "..self.shovelTriggerId);
			--print("Showing Plane of new FillType: "..tostring(self.heapPlanes[fillType].id).." with "..tostring(fillType));
			self.activePlaneId=self.heapPlanes[fillType];
			setVisibility(self.activePlaneId,true);
			self:forceMovePlane();
		end;
	end;

	-- change filltype
	if(fillType~=nil and fillType~=0)then
		if self.acceptedFillTypes[fillType]==true then
			self.fillType=fillType;
		end;
	else
		self.fillType=0;
	end;
	self:raiseDirtyFlags(self.placeableHeapDirtyFlag);
end;

function PlaceableHeap:getFillType()
	return self.fillType;
end;

function PlaceableHeap:allowFillType(fillTypeAdded)
	--print("Combine asked for "..tostring(fillTypeAdded));
	
	if(self.fillType==0 or self.fillType==fillTypeAdded)then
		return true;
	end;
	return false;
end;

function PlaceableHeap:getIsAttachedTo()
	return false;
end;

function PlaceableHeap:getAllowFillFromAir()
	--print("getAllowFillFromAir called");
	return self.allowFillFromAir;
end;

function PlaceableHeap:resetFillLevelIfNeeded(fillTypeCombine)
	if(fillTypeCombine ~= self.fillType)then
		self:setFillType(fillTypeCombine);
	end;
end;

function PlaceableHeap:tipTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	if g_currentMission.player ~= nil and otherActorId == g_currentMission.player.rootNode then
		if onEnter then
			self.playerInRange = true;
		else
			self.playerInRange = false;
		end;
	end;

	-- TipTrigger.triggerCallback
	if self.isEnabled then
		if onEnter then
			--BuyAtOncePlaceableHeap.isInsideHeap=true;
			do
				local trailer = g_currentMission.objectToTrailer[otherShapeId]
				if trailer ~= nil and trailer.allowTipDischarge then
					if g_currentMission.trailerTipTriggers[trailer] == nil then
						g_currentMission.trailerTipTriggers[trailer] = {}
					end
					table.insert(g_currentMission.trailerTipTriggers[trailer], self)
				end
			end
		elseif onLeave then
			--BuyAtOncePlaceableHeap.isInsideHeap=false;
			local trailer = g_currentMission.objectToTrailer[otherShapeId]
			if trailer ~= nil and trailer.allowTipDischarge then
				local triggers = g_currentMission.trailerTipTriggers[trailer]
				if triggers ~= nil then
					for i = 1, table.getn(triggers) do
						if triggers[i] == self then
							table.remove(triggers, i)
							if table.getn(triggers) == 0 then
								g_currentMission.trailerTipTriggers[trailer] = nil
							end
							break
						end
					end
				end
			end
		end
	end
end;

function PlaceableHeap:getTipInfoForTrailer(trailer, tipReferencePointIndex)
	local minDistance, bestPoint = self:getTipDistanceFromTrailer(trailer, tipReferencePointIndex);

	local isAllowed = false;

	local fillType=self.fillType;
	if(fillType==0)then
		if self.acceptedFillTypes[trailer.currentFillType] then
			isAllowed = true;
		end;
	elseif(fillType==trailer.currentFillType)then
		isAllowed = true;
	end;

	--print("TipInfoForTrailer: "..tostring(isAllowed)..", "..tostring(minDistance)..", "..tostring(bestPoint));
	--print("FillType in Heap is "..tostring(fillType).." and in Trailer "..tostring(trailer.currentFillType));
	return isAllowed, minDistance, bestPoint;
end;

function PlaceableHeap:updateTrailerTipping(trailer, fillDelta, fillType)
	--print("Fill Heap with "..tostring(fillDelta));
	if fillDelta < 0 then
		self:setFillType(fillType);
		self:setFillLevel(self.fillLevel-fillDelta);
	end;
end;

function PlaceableHeap:getTipDistance(trailerId)
	local trailerX, _, trailerZ = getWorldTranslation(trailerId);
	local triggerTipWidth=self.fillLevel/self.capacity*self.triggerTipWidth*1/3+self.triggerTipWidth*2/3;
	local distance=Utils.vector2Length(trailerX - self.pos.x, trailerZ - self.pos.z) - triggerTipWidth;
	if(distance<0)then distance=0; end;
	--print("FillLevel is "..self.fillLevel.." and triggerTipWidth "..triggerTipWidth);
	return distance;
end;

function PlaceableHeap:getTipDistanceFromTrailer(trailer, tipReferencePointIndex)
	local minDistance = math.huge
	local bestPoint;
	local bestWidth = -1;
	if tipReferencePointIndex ~= nil then
		minDistance=self:getTipDistance(trailer.tipReferencePoints[tipReferencePointIndex].node)
		bestPoint=tipReferencePointIndex;
	else
		for i, point in pairs(trailer.tipReferencePoints) do
			distance=self:getTipDistance(point.node);
			if point.width <= self.triggerTipWidth and bestWidth < point.width then
				if distance <= minDistance + 0.001 then
					bestPoint = i;
					bestWidth = point.width;
					minDistance = distance;
				end;
			elseif distance < minDistance then
				bestPoint = i;
				minDistance = distance;
			end;
		end;
	end;
	return minDistance, bestPoint
end;

registerPlaceableType("PlaceableHeap", PlaceableHeap);